from ..generic import *


class TruncatedGaussian1DAnalyticSpin(TruncatedGaussian1DAnalytic, SpinPopulationModel):
    pass

class TruncatedGaussian2DAnalyticSpin(TruncatedGaussian2DAnalytic, SpinPopulationModel):
    pass
